/***********************/
/* (c) 2007 dci@cpp.in */
/***********************/

#include "is.hpp"
#include "is-common.hpp"
#include "is-buffer.hpp"

/* buffer manager impl */
is_buffer::is_buffer ()
{
	this->dmax = 512;
	this->dlen = 0;
	this->data = xmalloc ( this->dmax );
}

is_buffer::~is_buffer ()
{
	xfree ( this->data );
}

void is_buffer::add ( size_t len )
{
	this->dlen += len;
	if ( this->dlen > this->dmax )
	{
		while ( this->dlen > this->dmax )
		{
			this->dmax *= 2;
		}
		this->data = xrealloc ( this->data, this->dmax );
	}
	memset ( ( ( char * ) this->data ) + this->dlen - len, 0, len );
}

size_t is_buffer::addb ( size_t len, void * idata )
{
	this->add ( len );
	if ( idata )
	{
		memcpy ( ( ( char * ) this->data ) + this->dlen - len, idata, len );
		return len;
	}
	else
	{
		return 0;
	}
}

size_t is_buffer::addc ( size_t len, is_buffer * ibuffer )
{
	this->add ( len );
	if ( ibuffer )
	{
		memcpy ( ( ( char * ) this->data ) + this->dlen - len, ibuffer->data, min ( len, ibuffer->dlen ) );
		return len;
	}
	else
	{
		return 0;
	}
}

void is_buffer::rem ( size_t len )
{
	size_t dbytes = min ( len, this->dlen );
	if ( dbytes )
	{
		memmove ( this->data, ( ( char * ) this->data ) + dbytes, this->dlen - dbytes );
		this->dlen -= dbytes;
		if ( this->dlen < this->dmax / 2 )
		{
			while ( this->dlen < this->dmax / 2 )
			{
				this->dmax /= 2;
			}
			this->data = xrealloc ( this->data, this->dmax );
		}
	}
}

size_t is_buffer::remb ( size_t len, void * obytes )
{
	size_t dbytes = min ( len, this->dlen );
	if ( obytes )
	{
		memcpy ( obytes, this->data, dbytes );
		this->rem ( dbytes );
		return dbytes;
		
	}
	else
	{
		this->rem ( dbytes );
		return 0;
	}
}

size_t is_buffer::remc ( size_t len, is_buffer * obuffer )
{
	size_t dbytes = min ( len, this->dlen );
	if ( obuffer )
	{
		obuffer->addb ( dbytes, this->data );
		this->rem ( dbytes );
		return dbytes;
		
	}
	else
	{
		this->rem ( dbytes );
		return 0;
	}
}

size_t is_buffer::len ()
{
	return this->dlen;
}
	
size_t is_buffer::read_data ( int offset, void * buf, size_t len )
{
	/* end-oriented? */
	if ( offset < 0 ) offset = this->dlen + offset;
	/* do! */
	if ( ( offset >= 0 ) && ( offset < this->dlen ) )
	{
		int dbytes = min ( len, this->dlen - offset );
		if ( dbytes )
		{
			memcpy ( buf, ( ( char * ) this->data ) + offset, dbytes );
		}
		return dbytes;
	}
	else
	{
		return 0;
	}
}

size_t is_buffer::write_data ( int offset, void * buf, size_t len )
{
	/* end-oriented? */
	if ( offset < 0 ) offset = this->dlen + offset;
	/* do! */
	if ( ( offset >= 0 ) && ( offset < this->dlen ) )
	{
		int dbytes = min ( len, this->dlen - offset );
		if ( dbytes )
		{
			memcpy ( ( ( char * ) this->data ) + offset, buf, dbytes );
		}
		return dbytes;
	}
	else
	{
		return 0;
	}
}

unsigned char is_buffer::read_byte ( int offset )
{
	/* end-oriented? */
	if ( offset < 0 ) offset = this->dlen + offset;
	/* do! */
	if ( ( offset >= 0 ) && ( ( offset + sizeof ( unsigned char ) ) <= this->dlen ) )
	{
		unsigned char * pt = ( ( unsigned char * ) this->data ) + offset;
		return pt[0];
	}
	else
	{
		return 0;
	}
}

size_t is_buffer::write_byte ( int offset, unsigned char byte )
{
	/* end-oriented? */
	if ( offset < 0 ) offset = this->dlen + offset;
	/* do! */
	if ( ( offset >= 0 ) && ( ( offset + sizeof ( unsigned char ) ) <= this->dlen ) )
	{
		unsigned char * pt = ( ( unsigned char * ) this->data ) + offset;
		pt[0] = byte;
		return sizeof ( unsigned char );
	}
	else
	{
		return 0;
	}
}

unsigned short is_buffer::read_le_word ( int offset )
{
	return ( ( this->read_byte ( offset + 1 ) << 8 ) | this->read_byte ( offset + 0 ) );
}

unsigned long is_buffer::read_le_dword ( int offset )
{
	return ( ( this->read_le_word ( offset + 2 ) << 16 ) | this->read_le_word ( offset ) );
}

size_t is_buffer::write_le_word ( int offset, unsigned short word )
{
	return ( write_byte ( offset + 0, ( word >> 0 ) & 0xFF ) + write_byte ( offset + 1, ( word >> 8 ) & 0xFF ) );
}

size_t is_buffer::write_le_dword ( int offset, unsigned long dword )
{
	return ( this->write_le_word ( offset + 0, ( dword >> 00 ) & 0xFFFF ) + this->write_le_word ( offset + 2, ( dword >> 16 ) & 0xFFFF ) );
}

unsigned short is_buffer::read_be_word ( int offset )
{
	return ( ( this->read_byte ( offset + 0 ) << 8 ) | this->read_byte ( offset + 1 ) );
}

unsigned long is_buffer::read_be_dword ( int offset )
{
	return ( ( this->read_be_word ( offset ) << 16 ) | read_be_word ( offset + 2 ) );
}

size_t is_buffer::write_be_word ( int offset, unsigned short word )
{
	return ( write_byte ( offset + 0, ( word >> 8 ) & 0xFF ) + write_byte ( offset + 1, ( word >> 0 ) & 0xFF ) );
}

size_t is_buffer::write_be_dword ( int offset, unsigned long dword )
{
	return ( this->write_be_word ( offset + 0, ( dword >> 16 ) & 0xFFFF ) + this->write_be_word ( offset + 2, ( dword >> 00 ) & 0xFFFF ) );
}
