/***********************/
/* (c) 2007 dci@cpp.in */
/***********************/

#include "is.hpp"
#include "is-common.hpp"

#include <iconv.h>

/* unaligned utf16 string meter */
size_t utf16len ( void * str )
{
	char * iptr = ( char * ) str;
	while ( iptr[0] || iptr[1] )
	{
		iptr += 2;
	}
	return ( ( ( size_t ) ( iptr - ( ( char * ) str ) ) ) / 2 );
}

/* unaligned utf16 string copier */
char * utf16dup ( void * str )
{
	size_t memsize = ( utf16len ( str ) + 1 ) * 2;
	char * ret = ( char * ) xmalloc ( memsize );
	memcpy ( ret, str, memsize );
	return ret;
}


/* mem manager stub */
void * xmalloc ( size_t size )
{
	void * result = malloc ( size );
	if ( result ) return result;
	vlog ( "xmalloc: Memory exhausted(trying to alloc %d bytes)", size );
	abort ( );
}

void * xrealloc ( void * ptr, size_t size )
{
	void * result = realloc ( ptr, size );
	if ( result ) return result;
	vlog ( "xrealloc: Memory exhausted(trying to alloc %d bytes)", size );
	abort ( );
}

void xfree ( void * ptr )
{
	free ( ptr );
}


/* iconv wrappers */
char * utf16_to_ascii ( char * utf16, int bytes )
{
	size_t   inlen = bytes;
	char *   in = utf16;
	size_t   outlen = inlen / 2;
	char *   out = ( char * ) xmalloc ( outlen + 1 );
	memset ( out, 0, outlen + 1 );
	/* save ptr */
	char * ret = out;
	/* convert to UTF-16 */
	iconv_t ic = iconv_open ( "ASCII", "UTF-16LE" );
	iconv ( ic, ( const char ** ) &in, ( size_t * ) &inlen, &out, ( size_t * ) &outlen );
	iconv_close ( ic );
	/* */
	* out = 0;
	return ret;
}

char * utf8_to_utf16 ( char * utf8, int bytes )
{
	size_t   inlen = bytes;
	char *   in = utf8;
	size_t   outlen = ( inlen + 2 ) * 2;
	char *   out = ( char * ) xmalloc ( outlen + 2 );
	memset ( out, 0, outlen + 2 );
	/* save ptr */
	char * ret = out;
	/* convert to UTF-16 */
	iconv_t ic = iconv_open ( "UTF-16LE", "UTF-8" );
	iconv ( ic, ( const char ** ) &in, ( size_t * ) &inlen, &out, ( size_t * ) &outlen );
	iconv_close ( ic );
	/* */
	out[0] = out[1] = 0;
	return ret;
}

char * cp1251_to_utf16 ( char * cp1251, int bytes )
{
	size_t inlen = bytes;
	char * in = cp1251;
	size_t outlen = ( inlen + 2 ) * 2;
	char * out = ( char * ) xmalloc ( outlen + 2 );
	memset ( out, 0, outlen + 2 );
	/* save ptr */
	char * ret = out;
	/* convert to UTF-16 */
	iconv_t ic = iconv_open ( "UTF-16LE", "CP1251" );
	iconv ( ic, ( const char ** ) &in, ( size_t * ) &inlen, &out, ( size_t * ) &outlen );
	iconv_close ( ic );
	/* */
	out[0] = out[1] = 0;
	return ret;
}
