/***********************/
/* (c) 2007 dci@cpp.in */
/***********************/

#include "is.hpp"

#include "is-proto-icq.hpp"

void icq_login ( is_proc * block )
{
	is_buffer packet;
	/* type */
	packet.add ( sizeof ( unsigned char ) );
	packet.write_byte ( 0, /* login */ 0x01 );
	/* packet data length. leave empty */
	packet.add ( sizeof ( unsigned long ) );
	/* maximum file size */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), MAX_FILE );
	/* auth */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), 0x00001f90 );
	/* FLAP magic */
	packet.add ( sizeof ( unsigned char ) );
	packet.write_byte ( - sizeof ( unsigned char ), 0x2a );
	/* FLAP channel */
	packet.add ( sizeof ( unsigned char ) );
	packet.write_byte ( - sizeof ( unsigned char ), 0x01 );
	/* FLAP seq(random) */
	packet.add ( sizeof ( unsigned short ) );
	packet.write_be_word ( - sizeof ( unsigned short ), time ( NULL ) & 0xFFFF );
	/* FLAP dsize */
	packet.add ( sizeof ( unsigned short ) );
	packet.write_be_word ( - sizeof ( unsigned short ), sizeof ( unsigned long ) + ( 2 * sizeof ( unsigned short ) ) + strlen ( block->is_login ) );
	/* proto version num = 1 */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_be_dword ( - sizeof ( unsigned long ), 0x00000001 );
	/* TLV type = uin */
	packet.add ( sizeof ( unsigned short ) );
	packet.write_be_word ( - sizeof ( unsigned short ), 0x0001 );
	/* TLV len */
	packet.add ( sizeof ( unsigned short ) );
	packet.write_be_word ( - sizeof ( unsigned short ), strlen ( block->is_login ) );
	/* TLV val */
	packet.addb ( strlen ( block->is_login ), block->is_login );
	/* write packet data len */
	packet.write_le_dword ( sizeof ( unsigned char ), packet.len ( ) - ( sizeof ( unsigned char ) + sizeof ( unsigned long ) ) );
	/* send */
	packet.remc ( packet.len ( ), block->buf_out );
}

void icq_raw ( is_proc * block )
{
	/* check, if raw packet received completely */
	if ( block->buf_in->len () >= ( block->buf_in->read_le_word ( 1 ) + 9 ) )
	{
		/* contactlist? */
		if ( ( block->buf_in->read_le_word ( 9 ) == 0x13 ) && ( block->buf_in->read_le_word ( 11 ) == 6 ) )
		{
			int i;
			unsigned short numrec = block->buf_in->read_be_word ( 0x14 );
			int roff = 0x16;
			
			/* load saved proto data */
			proto_data_icq * pdi = ( proto_data_icq * ) block->proto_data;
			/* empty? */
			if ( pdi == NULL )
			{
				/* new */
				pdi = ( proto_data_icq * ) xmalloc ( sizeof ( proto_data_icq ) );
				pdi->clist  = new is_clist;
				pdi->numrec = 0;
				/* save */
				block->proto_data = pdi;
			}
			/* update total records number */
			pdi->numrec += numrec;
						
			/* parse received */
			for ( i = 0; i < numrec; i++ )
			{
				/* name len */
				unsigned short nlen = block->buf_in->read_be_word ( roff );
				roff += 2;
				/* name(uin) */
				char * name;
				if ( nlen < 128 )
				{
					char * tbuf = ( char * ) xmalloc ( nlen );
					block->buf_in->read_data ( roff, tbuf, nlen );
					name = utf8_to_utf16 ( tbuf, nlen );
					xfree ( tbuf );
				}
				else
				{
					name = ( char * ) xmalloc ( 2 );
					name[0] = name[1] = 0;
				}
				roff += nlen;
				/* gid */
				unsigned short gid = block->buf_in->read_be_word ( roff );
				roff += 2;
				/* iid */
				unsigned short iid = block->buf_in->read_be_word ( roff );
				roff += 2;
				/* type */
				unsigned short type = block->buf_in->read_be_word ( roff );
				roff += 2;
				/* additional field len */
				unsigned short alen = block->buf_in->read_be_word ( roff );
				if ( alen > block->buf_in->len () )
				{
					alen = block->buf_in->len (); /* some sane value */
				}
				roff += 2;
				
				if ( type == 0x00 ) /* item */
				{
					/* reset fields */
					char * nick = utf16dup ( name );
					bool awaiting_auth = 0;
					/* read nick and authorization status from additional section */
					int aroff = roff;
					while ( aroff < roff + alen )
					{
						/* type */
						unsigned short type = block->buf_in->read_be_word ( aroff );
						aroff += 2;
						/* length */
						unsigned short len  = block->buf_in->read_be_word ( aroff );
						if ( len > ( ( roff + alen ) - aroff ) )
						{
							len = ( roff + alen ) - aroff;
						}
						aroff += 2;
						/* process value */
						if ( type == 0x0131 ) /* nick */
						{
							xfree ( nick );
							char * tbuf = ( char * ) xmalloc ( len );
							block->buf_in->read_data ( aroff, tbuf, len );
							nick = utf8_to_utf16 ( tbuf, len );
							xfree ( tbuf );
						}
						else if ( type == 0x0066 ) /* awaiting_authorization */
						{
							awaiting_auth = 1;
						}
						/* finish */
						aroff += len;
					}
					/* add item */
					pdi->clist->addi ( gid, iid, awaiting_auth, 0, name, nick );
					/* */
					xfree ( nick );
				}
				else if ( type == 0x01 ) /* group */
				{
					/* add group */
					if ( gid > 0 )
					{
						pdi->clist->addg ( gid, name );
					}
				}
				roff += alen;
				/* */
				xfree ( name );
			}
			/* contactlist finished? */
			if ( ( block->buf_in->read_be_word ( 13 ) & 0x01 ) == 0 )
			{
				/* create out packet */
				is_buffer packet;
				/* type */
				packet.add ( sizeof ( unsigned char ) );
				packet.write_byte ( 0, /* raw */ 2 );
				/* length, skip value for now */
				packet.add ( sizeof ( unsigned long ) );
				


				/* identity block. following 11 bytes can contain any value, some can be 0 */
				/* list update time */
				time_t lutime = 0;
				if ( ( roff + sizeof ( unsigned long ) ) <= ( block->buf_in->read_le_word ( 1 ) + 9 ) )
				{
					/* are there one more dword in packet? that's time! */
					lutime = block->buf_in->read_be_dword ( roff );
					roff += sizeof ( unsigned long );
				}
				packet.add ( sizeof ( unsigned long ) );
				packet.write_le_dword ( - sizeof ( unsigned long ), lutime );
				/* total processed records number */
				packet.add ( sizeof ( unsigned short ) );
				packet.write_le_word ( - sizeof ( unsigned short ), pdi->numrec );
				/* unk1. some hash */
				packet.addb ( 2, ( char * ) "ab" );
				/* unk2. 4 seems to be default for icq */
				packet.add ( sizeof ( unsigned char ) );
				packet.write_byte ( - sizeof ( unsigned char ), 4 );
				/* unk3. some hash */
				packet.addb ( 2, ( char * ) "yz" );
			
				
				
				/* write contactlist */
				pdi->clist->dump ( &packet );
				/* write length value to packet */
				packet.write_le_dword ( sizeof ( unsigned char ), packet.len ( ) );
				/* send packet */
				packet.remc ( packet.len (), block->buf_out );
				/* free */
				delete ( pdi->clist );
				xfree ( pdi );
				block->proto_data = NULL;
			}
		}
		/* discard block */
		block->buf_in->rem ( block->buf_in->read_le_word ( 1 ) + 9 );
	}
}

void icq_quit ( is_proc * block )
{
	block->buf_in->rem ( block->buf_in->len ( ) );
}

is_proto_t proto_icq = { "icq", icq_login, icq_raw, icq_quit };
