/***********************/
/* (c) 2007 dci@cpp.in */
/***********************/

#include "is.hpp"

#include "is-proto-mrim.hpp"

/* used to skip unknown structmask fields */
size_t mrim_skipmask ( is_buffer * buf, size_t offset, char * mask )
{
	while ( * mask )
	{
		if ( * mask == 'u' ) /* dword */
		{
			offset += 4;
		}
		else if ( * mask == 's' ) /* declared-length string */
		{
			int slen = buf->read_le_dword ( offset );
			offset += ( 4 + slen );
		}
		else if ( * mask == 'z' ) /* sz */
		{
			while ( buf->read_byte ( offset ) )
			{
				offset++;
			}
			offset++; /* skip zero */
		}
		mask++;
	}
	return offset;
}

void mrim_login ( is_proc * block )
{
	is_buffer packet;
	/* type */
	packet.add ( sizeof ( unsigned char ) );
	packet.write_byte ( 0, /* login */ 0x01 );
	/* packet data length. leave empty */
	packet.add ( sizeof ( unsigned long ) );
	/* maximum file size */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), MAX_FILE );
	/* auth */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), 0x00001f90 );
	/* mrim magic */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), 0xDEADBEEF );
	/* mrim proto ver */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), 0x00010007 );
	/* mrim seq */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), 0x00000002 );
	/* mrim msg */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), /* MRIM_CS_LOGIN2 */ 0x00001038 );
	/* mrim dlen */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), /* login len */ sizeof ( unsigned long ) + strlen ( block->is_login ) );
	/* mrim from */
	packet.add ( sizeof ( unsigned long ) );
	/* mrim fromport */
	packet.add ( sizeof ( unsigned long ) );
	/* mrim reserved */
	packet.add ( 16 * sizeof ( unsigned char ) );
	/* login len */
	packet.add ( sizeof ( unsigned long ) );
	packet.write_le_dword ( - sizeof ( unsigned long ), strlen ( block->is_login ) );
	/* login */
	packet.addb ( strlen ( block->is_login ), block->is_login );
	/* write packet data len */
	packet.write_le_dword ( sizeof ( unsigned char ), packet.len ( ) - ( sizeof ( unsigned char ) + sizeof ( unsigned long ) ) );
	/* send */
	packet.remc ( packet.len ( ), block->buf_out );
}

void mrim_raw ( is_proc * block )
{
	/* check, if raw packet received completely */
	if ( block->buf_in->len () >= ( block->buf_in->read_le_word ( 1 ) + 5 ) )
	{
		/* contactlist? */
		if ( block->buf_in->read_le_dword ( 0x11 ) == /* MRIM_CS_CONTACT_LIST2 */ 0x1037 )
		{
			/* mrim data and len */
			unsigned dlen = block->buf_in->read_le_dword ( 0x15 );
			int roff = 0x31;
			/* parse mrim data */
			if ( block->buf_in->read_le_dword ( roff ) == /* GET_CONTACTS_OK */ 0x0000 )
			{
				int i;
				/* create list */
				is_clist contactlist;
				/* read! */
				roff += sizeof ( unsigned long );
				/* first N entries = groups */
				unsigned long group_num = block->buf_in->read_le_dword ( roff );
				roff += sizeof ( unsigned long );
				/* groupmask length */
				int gmlen = block->buf_in->read_le_dword ( roff );
				roff += sizeof ( unsigned long );
				/* groupmask */
				char groupmask[64];
				memset ( groupmask, 0, sizeof ( groupmask ) );
				block->buf_in->read_data ( roff, groupmask, min ( gmlen, sizeof ( groupmask ) - 1 ) );
				roff += gmlen;
				/* itemmask length */
				int imlen = block->buf_in->read_le_dword ( roff );
				roff += sizeof ( unsigned long );
				/* itemmask */
				char itemmask[64];
				memset ( itemmask, 0, sizeof ( itemmask ) );
				block->buf_in->read_data ( roff, itemmask, min ( imlen, sizeof ( itemmask ) - 1 ) );
				roff += imlen;
				
				/* read groups */
				for ( i = 0; i < group_num; i++ )
				{
					/* gid(most significant byte of flags), inlux skips gid=0, so + 1 */
					unsigned short gid = ( unsigned short ) ( block->buf_in->read_le_dword ( roff ) >> 24 ) + 1;
					roff += sizeof ( unsigned long );
					/* name length */
					unsigned long nlen = block->buf_in->read_le_dword ( roff );
					if ( nlen > 256 )
					{
						nlen = 256;					
					}
					roff += sizeof ( unsigned long );
					/* name */
					char * tdata = ( char * ) xmalloc ( nlen );
					block->buf_in->read_data ( roff, tdata, nlen );
					char * name = cp1251_to_utf16 ( tdata, nlen );
					xfree ( tdata );
					roff += nlen;
					/* skip all pased files, extenting 2 read */
					roff = mrim_skipmask ( block->buf_in, roff, groupmask + 2 );
					/* add group */
					contactlist.addg ( gid, name );
				}
				
				i = 0; /* used for item id simulation */
				/* read items */
				while ( roff < dlen )
				{
					/* read contact flags */
					unsigned long flags = block->buf_in->read_le_word ( roff );
					roff += sizeof ( unsigned long );
					if ( flags & 0x21 /* CONTACT_FLAG_REMOVED | CONTACT_FLAG_SHADOW */ )
					{
						/* skip all, except for read flags */
						roff = mrim_skipmask ( block->buf_in, roff, itemmask + 1 );
					}
					else
					{
						/* read gid, extend by 1 */
						unsigned short gid = ( unsigned short ) block->buf_in->read_le_dword ( roff ) + 1;
						roff += sizeof ( unsigned long );
						/* name len */
						unsigned long nlen = block->buf_in->read_le_dword ( roff );
						if ( nlen > 256 )
						{
							nlen = 256;					
						}
						roff += sizeof ( unsigned long );
						/* name */
						char * tdata = ( char * ) xmalloc ( nlen );
						block->buf_in->read_data ( roff, tdata, nlen );
						char * name = cp1251_to_utf16 ( tdata, nlen );
						xfree ( tdata );
						roff += nlen;
						/* nick len */
						nlen = block->buf_in->read_le_dword ( roff );
						if ( nlen > 256 )
						{
							nlen = 256;					
						}
						roff += sizeof ( unsigned long );
						/* nick */
						tdata = ( char * ) xmalloc ( nlen );
						block->buf_in->read_data ( roff, tdata, nlen );
						char * nick = cp1251_to_utf16 ( tdata, nlen );
						xfree ( tdata );
						roff += nlen;
						/* server flags */
						unsigned long sflags = block->buf_in->read_le_dword ( roff );
						roff += sizeof ( unsigned long );
						/* contactstatus */
						unsigned long cstatus = block->buf_in->read_le_dword ( roff );
						roff += sizeof ( unsigned long );
						/* skip mask fields, excluding read 6 */
						roff = mrim_skipmask ( block->buf_in, roff, itemmask + 6 );
						/* fix status, is not authorized */
						if ( sflags & 0x01 /* CONTACT_INTFLAG_NOT_AUTHORIZED */ )
						{
							cstatus = 8; /* inlux 'awaiting authorization' */
						}
						/* add item */
						contactlist.addi ( gid, i, sflags & 0x01, cstatus, name, nick );
						/* next simulated item */
						i++;
					}
				}
				
				/* send cl packet */
				is_buffer packet;
				/* type */
				packet.add ( sizeof ( unsigned char ) );
				packet.write_byte ( 0, /* raw */ 2 );
				/* length, skip value for now */
				packet.add ( sizeof ( unsigned long ) );
				/* list update time */
				packet.add ( sizeof ( unsigned long ) );
				/* ss */
				packet.addb ( 7, ( char * ) "abcdefg" );
				/* write contactlist */
				contactlist.dump ( &packet );
				/* write length value to packet */
				packet.write_le_dword ( sizeof ( unsigned char ), packet.len ( ) );
				/* send packet */
				packet.remc ( packet.len (), block->buf_out );
			}
		}
		/* discard block */
		block->buf_in->rem ( block->buf_in->read_le_word ( 1 ) + 5 );
	}
}

void mrim_quit ( is_proc * block )
{
	block->buf_in->rem ( block->buf_in->len ( ) );
}

is_proto_t proto_mrim = { "mail.ru", mrim_login, mrim_raw, mrim_quit };
