/***********************/
/* (c) 2007 dci@cpp.in */
/***********************/

#include "is.hpp"
#include "is-server.hpp"

/* KE globals */
kevent_t * ke_vec = NULL;
unsigned   ke_vec_alloc = 0;
unsigned   ke_vec_used = 0;

/* kqueue updater */
void ke_change ( int ident, int filter, int flags, int data, void * is_block )
{
	/* check if allocated memory exhausted */
	if ( ke_vec_used == ke_vec_alloc )
	{
		ke_vec_alloc *= 2;
		ke_vec = ( kevent_t * ) xrealloc ( ke_vec, ke_vec_alloc * sizeof ( * ke_vec ) );
	}
	/* fill new item */
	ke_vec[ke_vec_used].ident = ident;
	ke_vec[ke_vec_used].filter = filter;
	ke_vec[ke_vec_used].flags = flags;
	ke_vec[ke_vec_used].fflags = 0;
	ke_vec[ke_vec_used].data = data;
	ke_vec[ke_vec_used].udata = is_block;
	/* move up */
	ke_vec_used++;
}

/* kqueue executive */
void ke_loop ( int listen_sock )
{
	/* create kqueue */
	int kq = kqueue ();
	if ( kq != -1 )
	{
		/* alloc mem for changes list */
		ke_vec_alloc = 2 * INIT_USERS; /* 2 events for user */
		ke_vec = ( kevent_t * ) xmalloc ( ke_vec_alloc * sizeof ( * ke_vec ) );
		/* add acceptor */
		ke_change ( listen_sock, EVFILT_READ, EV_ADD | EV_ENABLE, 0, NULL );
		/* loop */
		while ( "punk not dead" )
		{
			/* push changes/retrieve events */
			int n = kevent ( kq, /* in */ke_vec, ke_vec_used, /* out */ke_vec, ke_vec_alloc, NULL );
			/* parse result */
			if ( n > 0 )
			{
				kevent_t * kep;
				/* copy event queue */
				kevent_t * revents = ( kevent_t * ) xmalloc ( n * sizeof ( kevent_t ) );
				memcpy ( revents, ke_vec, n * sizeof ( kevent_t ) );
				/* set queue empty for further modification */
				ke_vec_used = 0;
				/* process queue data */
				for ( kep = revents; kep < revents + n; kep++)
				{
					if ( kep->filter == EVFILT_READ ) /* connect or data arrived */
					{
						if ( kep->ident == listen_sock ) /* connect */
						{
							sockaddr_in sin;
							socklen_t sinsize = sizeof ( sin );
							int s = accept ( kep->ident, ( struct sockaddr * ) &sin, &sinsize );
							if ( s != -1 )
							{
								/* set socket to keepalive state */
								int one = 1;
								setsockopt ( s, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof ( one ) );
								/* in linux also manual keepalive settings should be passed via TCP_KEEPIDLE, TCP_KEEPINTVL and TCP_KEEPCNT with setsockopt.
								   in freebsd it should be set via sysctl with is_sysset.sh */
								   
								/* create new processor block */
								is_proc * newblock = new is_proc ( s, sin.sin_addr, sin.sin_port );
								/* set events */
								ke_change ( s, EVFILT_READ, EV_ADD | EV_ENABLE, 0, newblock );
								ke_change ( s, EVFILT_WRITE, EV_ADD | EV_DISABLE, 0, newblock );
							}
							else
							{
								vlog ( "accept(): error %d \"%s\"", errno, strerror ( errno ) );
								break;
							}
						}
						else /* data arrived */
						{
							is_proc * block = ( is_proc * ) kep->udata;
							if ( ( /* data length */ kep->data >= 0 ) && block )
							{
								int n;
								/* read data */
								void * tdata = xmalloc ( /* data length */ kep->data );
								n = recv ( block->sk_fd, tdata, /* data length */ kep->data, 0 );
								if ( n > 0 )
								{
									/* put to buffer */
									block->buf_in->addb ( /* data length */ kep->data, tdata );
									/* call parse executive */
									block->proc_do ( );
									/* check if socket need to be switched to send */
									if ( block->buf_out->len () > 0 )
									{
										ke_change ( kep->ident, EVFILT_READ, EV_DISABLE, 0, block );
										ke_change ( kep->ident, EVFILT_WRITE, EV_ENABLE, 0, block );
									}
								}
								else
								{
									/* explicitly delete events from kqueue */
									ke_change ( kep->ident, EVFILT_READ, EV_DELETE, 0, NULL );
									ke_change ( kep->ident, EVFILT_WRITE, EV_DELETE, 0, NULL );
									/* destroy processor block */
									delete block;
									/* close endpoint */
									close ( kep->ident );
								}
								/* free recv buffer */
								free ( tdata );
							}
						}
					}
					else if ( kep->filter == EVFILT_WRITE ) /* data ready to write */
					{
						is_proc * block = ( is_proc * ) kep->udata;
						if ( block )
						{
							/* how many bytes we are going to write? */
							size_t numbytes = min ( block->buf_out->len (), /* buffer len */ kep->data );
							if ( numbytes > 0 )
							{
								/* read data */
								void * tdata = xmalloc ( numbytes );
								block->buf_out->remb ( numbytes, tdata );
								/* send data */
								send ( block->sk_fd, tdata, numbytes, 0 );
								/* free temp */
								xfree ( tdata );
							}
							/* check if socket need to be switched to recv */
							if ( block->buf_out->len() == 0 )
							{
								ke_change ( kep->ident, EVFILT_WRITE, EV_DISABLE, 0, block );
								ke_change ( kep->ident, EVFILT_READ, EV_ENABLE, 0, block );
							}
						}
					}
				}
				/* free copied events */
				xfree ( revents );
			}
			else
			{
				vlog ( "kevent()=%d: error %d \"%s\"", n, errno, strerror ( errno ) );
				break;
			}
		}
		/* free changes list mem */
		ke_vec_alloc = ke_vec_used = 0;
		xfree ( ke_vec );
		ke_vec = NULL;
		/* close kqueue */
		close ( kq );
	}
	else
	{
		vlog ( "kqueue(): error %d \"%s\"", errno, strerror ( errno ) );
	}
}

/* main server proc */
void server ( char * iface, char * root, uid_t user, gid_t group )
{
	int server_sock = socket ( PF_INET, SOCK_STREAM, 0 );
	if ( server_sock != -1 )
	{
		/* reuse for listening */
		int one = 1;
		setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof one);
	  	/* create endpoint */
	  	sockaddr_in sin;
		memset ( &sin, 0, sizeof ( sin ) );
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = iface ? inet_addr ( iface ) : 0;
		sin.sin_port = htons ( 8090 );
		if ( bind ( server_sock, ( const struct sockaddr * ) &sin, sizeof ( sin ) ) != -1)
		{
			if ( listen ( server_sock, 255 ) != -1 )
			{
				/* some sec stuff */
				if ( root )
				{
					chroot ( root );
				}
				if ( ( group || user ) && ( 0 == getuid ( ) ) )
				{
					setgid ( group );
					setuid ( user );
				}
				/* notify user */
				vlog ( "[%08X] Listening...", server_sock );
				/* run queue mech */
				ke_loop ( server_sock );
			}
			else
			{
				vlog ( "listen(): error %d \"%s\"", errno, strerror ( errno ) );
			}
		}
		else
		{
			vlog ( "bind(): error %d \"%s\"", errno, strerror ( errno ) );
		}
		shutdown ( server_sock, SHUT_RDWR );
		close ( server_sock );
	}
	else
	{
		vlog ( "socket(): error %d \"%s\"", errno, strerror ( errno ) );
	}
}
